"""Sample accelerator phasespace and detector volume

"""

__author__ = 'tunnell'

import math
import numpy as np
from Constants import MUON_MASS


class Sampler(object):
    """Sample"""

    def __init__(self, E_mu, baseline):
        """Setup the accelerator and detector phasespace calculator

        :param E_mu: central muon energy in decay ring (GeV)
        :param baseline: distance between accelerator and detector (km)

        The baseline should be the
        distance between the accelerator and the detector.  More specifically,
        it's the point between the most downstream part of the accelerator and
        the most upstream part of the detector.  This is the minimum distance
        that a neutrino could possibly travel between the accelerator and
        detector.
        """
        self.detector_radius = 0.0  # km
        self.detector_mass = 0.0  # grams
        self.detector_density = 0.0  # g/km^3

        self.accelerator_straight_length = 0.0  # km
        self.E_mu = E_mu  # GeV
        self.E_mu_spread = 0.0  # GeV

        self.baseline = baseline  # km

        self.emittance = None
        self.beta = None

    def set_accelerator_twiss(self, emittance, beta):
        """Set twiss parameters

        Alpha (ie. beta-prime) is assumed to be zero.
        These are the central momentum optics.
        """
        self.emittance = emittance
        self.beta = beta

    def set_accelerator_length(self, accelerator_straight_length):
        """set length in km
        """
        self.accelerator_straight_length = accelerator_straight_length

    def set_accelerator_energy(self, energy=None, spread=None):
        if energy:
            self.E_mu = energy
        if spread:
            self.E_mu_spread = spread

    def set_detector(self, radius, mass, density):
        """Set detector volume parameters

        :param radius: detector radius in m
        :param mass: detector mass in kt
        :param density: density of detector in g/cm^3

        :returns: nothing
        """

        radius, mass, density = [float(x) for x in [radius, mass, density]]

        # Convert: m -> km
        self.detector_radius = radius / 1000

        # Convert: kt -> grams
        self.detector_mass = mass * 10 ** 9

        # We do a units conversion here from g/cm^3 to g/km^3
        self.detector_density = density * 10 ** 15

        if self.detector_radius:
            self.detector_max_z = self.detector_mass
            self.detector_max_z /= self.detector_density
            self.detector_max_z /= (2 * self.detector_radius) ** 2

            #print 'max_detector_z', self.detector_max_z

    def sample_accelerator(self, nbins):
        """Sample a point in the accelerator phasespace

        a is x, km
        b is x', radians
        c is uniform value for z in straight, km
        """

        if self.emittance and self.beta:
            # From Penn 61 MUCOOL note
            mean = [0, 0]
            cov = [[0, 0], [0, 0]]  # covariance matrix

            #  Ignore cross terms (done in initialization)
            #ie. cov[0][1] = 0.0 # sigma_xx'
            #ie. cov[0][1] = 0.0 # sigma_x'x

            # sigma_xx
            cov[0][0] = self.emittance * self.beta * MUON_MASS / self.E_mu

            # sigma_x'x'
            cov[1][1] = MUON_MASS * self.emittance / self.beta / self.E_mu

            # Step 1: compute phasespace
            result = np.random.multivariate_normal(mean, cov, nbins)
            result = result.astype(np.float32)
            a, b = result.T  # transpose for phasespace x, x-prime
        else:
            a, b = np.zeros([2, nbins], dtype=np.float32)

        #  This is used for figuring out where along the ring the muon decays
        c = np.random.uniform(high=self.accelerator_straight_length,
            size=nbins).astype(np.float32)

        return a, b, c

    def sample_detector(self, nbins):
        """Sample detector volume

        d is uniform values for x,y,z in detector
        """

        #  Where in the detector volume the decay happens?
        #  The size is 3*nbins because there are three spacial dimensions
        if self.detector_radius == 0.0:
            d = np.zeros(3 * nbins)
        else:
            #  In compute final states, we multiply by length in x, y, z
            d = np.random.uniform(
                low=-1, high=1,
                size=3 * nbins)
            d = d.astype(np.float32)

        return d

    def compute_final_states(self, N, a, b, c, d):
        """Compute distances and angles"""
        multiplier = 3
        for i in range(N):
            z = 0
            r = 0
            if self.detector_radius > 0:
                x = self.detector_radius * d[multiplier * i]
                y = self.detector_radius * d[multiplier * i + 1]

                z = self.detector_max_z * (d[multiplier * i + 2] + 1) / 2

                r = math.sqrt(x * x + y * y)

            radius = a[i]

            distance = c[i]  # somewhere along the accelerator straight
            distance += self.baseline
            distance += z

            distance = math.sqrt(distance * distance + (radius - r) ** 2)

            angle_lab = b[i]
            angle_lab = math.asin(r / distance) - angle_lab

            a[i] = distance
            b[i] = angle_lab

        return [a, b]

    def sample(self, N=1):
        """Sample phasespace
        """

        a, b, c = self.sample_accelerator(N)

        d = self.sample_detector(N)

        distance, angle = self.compute_final_states(N, a, b, c, d)

        E = np.random.uniform(low=(self.E_mu - self.E_mu_spread),
            high=(self.E_mu + self.E_mu_spread),
            size=N)
        return zip(E, distance, angle)



__author__ = 'tunnell'


def print_samples_to_screen(results):
    for E, l, angle in results:
        print 'E_mu: %g GeV\tL: %g km\tangle:%g rad' % (E, l, angle)


def print_events_to_screen(events):
    print 'Total number of events:'
    for key, value in events.iteritems():
        flavor, polarity = key

        print '\t', flavor, polarity, value


def create_modglobes_samples_file(samples, N_mu):
    file_obj = open('samples', 'w')
    file_obj.write('%d\n' % len(samples))
    file_obj.write('%g\n' % N_mu)
    for E, l, angle in samples:
        file_obj.write('%g %g %g\n' % (E, l, angle))
    file_obj.close()



def create_plot(name, results, use_matplotlib=True):
    """Create plot with matplotlib"""

    try:
        import matplotlib.pyplot as plt
    except:
        print "Can't find matplotlib.  Just creating flux output."
        use_matplotlib = False

    for key in results.keys():
        plt.figure()
        par = results[key]

        x = []
        y = []
        for key2, val in par.iteritems():
            x.append(key2)
            y.append(val)

        if use_matplotlib:
            plt.plot(x, y, 'ro')
            plt.xlabel("Neutrino energy [GeV]")
            plt.ylabel(r'Flux /(m${}^2$ GeV)')

            plt.savefig('avg_%s_%s.eps' % (name, key))

        x.sort()

        file_obj = open('raw_%s_%s' % (name, key), 'w')
        for sorted_x in x:
            file_obj.write('%f %f\n' % (sorted_x, par[sorted_x]))
        file_obj.close()

        print key, 'sum', sum(y)



__author__ = 'tunnell'

from Constants import AVOGADROS_NUMBER

def CC(fluxes, mass=1):
    """Compute CC event rates

    mass in kt
    """

    xsec = {}
    xsec['nubar'] = 0.34 * 10 ** -42 # units of m^2
    xsec['nu'] = 0.67 * 10 ** -42     # units of m^2

    E_range = {'min': None, 'max': None}
    bins = {}

    interactions = {}
    for flavor, value in fluxes.iteritems():
        for energy, flux in value.iteritems():
            #  Units are flux/GeV, so need to know GeV range
            if E_range['min'] is None or energy < E_range['min']:
                E_range['min'] = energy
            if E_range['max'] is None or energy > E_range['max']:
                E_range['max'] = energy

            for nu_or_nubar in ['nu', 'nubar']:
                key = (flavor, nu_or_nubar)
                if key not in interactions:
                    interactions[key] = 0.0
                    bins[key] = 0

                my_int = mass * 10 ** 9 * AVOGADROS_NUMBER * xsec[nu_or_nubar]
                my_int *= energy * flux

                interactions[key] += my_int
                bins[key] += 1

    return interactions





"""Generate nuSTORM near detector fluxes

For details, see generate-nustorm-fd-flux.py which is commented."""

__author__ = 'tunnell'

from msrflux.Sampler import Sampler
import msrflux.Output as Output
from msrflux.DecayKinematics import compute_neutrino_flux
from msrflux.InteractionRate import CC

energy_muon = 3.8  # GeV
baseline = 0.1  # km

x = Sampler(energy_muon, baseline)

#x.set_accelerator_energy(spread=0.1 * energy_muon) # GeV

my_mass = 0.1 # kt
x.set_detector(radius=1.0, # m
    mass=my_mass, # kt
    density=1.4)  # g/cm^3

x.set_accelerator_length(0.15) # km

results = x.sample(N=1000)
N_mu = 1.8e18
Output.create_modglobes_samples_file(results, N_mu)
results = compute_neutrino_flux(results, N_mu=1.8e18) # 1/m2

Output.create_plot('nustorm_near', results)
Output.print_events_to_screen(CC(results, mass=my_mass))



"""Double differential cross section"""

__author__ = 'tunnell'

import math
import numpy as np
from Constants import MUON_MASS


def compute_neutrino_flux(my_input, N_mu=1.8e18, E_min=0.5, E_max=3.8,
                          steps=100):
    """Compute neutrino fluxes using double differential cross section

    We assume unpolarized.

    For a good reference of the equation for neutrino fluxes, see
    hep-ph/0002108.  For a derivation, see Jian Tang's thesis.
    """

    nu_energies = np.linspace(E_min, E_max, steps)
    N_flux_bins = len(my_input)

    results = {'e_flux': {}, 'mu_flux': {}}

    for E_nu in nu_energies:
        E_nu = float(E_nu)
        results['e_flux'][E_nu] = []
        results['mu_flux'][E_nu] = []

    for event in my_input:
        E_mu, distance, angle = event
        # E_mu in GeV
        # distance in km
        # angle in rad

        distance = 1000 * distance  # km -> m

        for E_nu in nu_energies:
            E_nu = float(E_nu)

            y = E_nu / E_mu  # unitless, (0,1)

            if y < 0 or y > 1:
                continue

            beta = math.sqrt(1 - MUON_MASS ** 2 / E_mu ** 2)

            flux_e = 24.0 * N_mu
            flux_e /= (math.pi * distance ** 2 * MUON_MASS ** 6)
            flux_e *= E_mu ** 4 * y ** 2
            flux_e *= (1 - beta * math.cos(angle))
            flux_e *= (MUON_MASS ** 2 - 2 * E_mu ** 2 *
                                        y * (1 - beta * math.cos(angle)))

            flux_mu = 4.0 * N_mu
            flux_mu /= (math.pi * distance ** 2 * MUON_MASS ** 6)
            flux_mu *= E_mu ** 4 * y ** 2
            flux_mu *= (1 - beta * math.cos(angle))
            flux_mu *= (3 * MUON_MASS ** 2 - 4 * E_mu ** 2 *
                                             y * (1 - beta * math.cos(angle)))

            # Fluxes go negative if unphysical
            if flux_e > 0.0:
                results['e_flux'][E_nu].append(flux_e)
            if flux_mu > 0.0:
                results['mu_flux'][E_nu].append(flux_mu)

    for flux_name in results:
        for E_nu in results[flux_name]:
            my_sum = sum(results[flux_name][E_nu])
            my_sum /= N_flux_bins
            my_sum /= len(nu_energies)

            results[flux_name][E_nu] = my_sum

    return results



"""Compute nustorm FD neutrino fluxes"""

__author__ = 'tunnell'

from msrflux.Sampler import Sampler
import msrflux.Output as Output
from msrflux.DecayKinematics import compute_neutrino_flux
from msrflux.InteractionRate import CC

#  What is the mean stored muon energy in the decay ring?
energy_muon = 3.8  # GeV

#  Set the baseline between the accelerator and detector.
baseline = 2.0  # km

#  Sampler performs the MC integration by sampling points in the detector and
#  accelerator volumes.
x = Sampler(energy_muon, baseline)

###########################
##  OPTIONAL SET METHODS ##
###########################

# Smear the stored muon energy by 10% uniform spread
#x.set_accelerator_energy(spread=0.1 * energy_muon) # GeV

#  Integrate over the detector volume
my_mass = 1.3 # kt
x.set_detector(radius=2.5, # m
    mass=my_mass, # kt
    density=7.8)  # g/cm^3

#  Integrate over the detector straight
x.set_accelerator_length(0.15) # km

#  Set the Twiss parameters (ie. accelerator optics).  This doesn't matter
#  much but is here for complessness.
x.set_accelerator_twiss(emittance=2.1e-06,
    beta=0.04) # km, km

#  Sample N points
results = x.sample(N=1000)

# Print to screen
Output.print_samples_to_screen(results)

# Compute the fluxes
results = compute_neutrino_flux(results, N_mu=1.8e18) # 1/m2

#  Create plots and flux files
Output.create_plot('nustorm', results)

Output.print_events_to_screen(CC(results, mass=my_mass))



from distutils.core import setup

setup(
    name='msrflux',
    version='0.3.1',
    packages=['msrflux'],
    scripts=['bin/generate-nustorm-fd-flux.py',
             'bin/generate-nustorm-nd-flux.py',
             'bin/generate-henf-flux.py'],
    url='http://pypi.python.org/pypi/msrflux/',
    license='LICENSE.txt',
    author='tunnell',
    author_email='c.tunnell1@physics.ox.ac.uk',
    description='Computes the neutrino flux from a muon decay ring while averaging over the detector and accelerator.',
    long_description=open('README.txt').read(),
    install_requires=[
        "numpy >= 1.6.0",
    ]
)



"""Compute nustorm FD neutrino fluxes"""

__author__ = 'tunnell'

from msrflux.Sampler import Sampler
import msrflux.Output as Output
from msrflux.DecayKinematics import compute_neutrino_flux
from msrflux.InteractionRate import CC

energy_muon = 50.0  # GeV
baseline = 3000.0  # km
N_mu = 10.66e+20 * 4 # muons
my_mass = 50.0 # kt 

x = Sampler(energy_muon, baseline)

results = x.sample(N=100)

Output.print_samples_to_screen(results)

results = compute_neutrino_flux(results, N_mu=N_mu,
    E_min=4.0, E_max=energy_muon) # 1/m2

Output.create_plot('henf', results)
Output.print_events_to_screen(CC(results, mass=my_mass))
